import * as THREE from '../build/three.module.js';
import Stats from './jsm/libs/stats.module.js';
import { GUI } from './jsm/libs/dat.gui.module.js';
import { OrbitControls } from './jsm/controls/OrbitControls.js';
import { Water } from './jsm/objects/Water.js';
import { Sky } from './jsm/objects/Sky.js';

class App {
    constructor() {
        const domWebGL = document.createElement('div');
        document.body.appendChild(domWebGL);
        this._domWebGL = domWebGL;

        const renderer = new THREE.WebGLRenderer({ antialias: true });
        renderer.outputEncoding = THREE.sRGBEncoding;
        renderer.setPixelRatio(window.devicePixelRatio);
        //renderer.shadowMap.enabled =  true;
        renderer.toneMapping = THREE.ACESFilmicToneMapping;
        domWebGL.appendChild(renderer.domElement);
        this._renderer = renderer;

        const scene = new THREE.Scene();
        this._scene = scene;

        this._setupCamera();
        //this._setupLight();
        //this._setupPostprocessing();
        //this._setupInteractions();
        this._setupModel();
        this._setupGUI();
        this._setupControls();
        //this._setupHelper();

        window.onresize = this.resize.bind(this);

        this.resize();
        this.render()
    }

    _setupInteractions() {
    }

    _setupPostprocessing() {
        //.
    }

    _updateSun() {
        const sun = new THREE.Vector3();
        const phi = THREE.MathUtils.degToRad(90 - this._parameters.elevation);
        const theta = THREE.MathUtils.degToRad(this._parameters.azimuth);

        sun.setFromSphericalCoords(1, phi, theta);
        this._sky.material.uniforms["sunPosition"].value.copy(sun);
        this._water.material.uniforms["sunDirection"].value.copy(sun).normalize();

        this._scene.environment = new THREE.PMREMGenerator(this._renderer).fromScene(this._sky).texture;
    }

    _setupModel() {
        const waterGeometry = new THREE.PlaneGeometry(10000, 10000);
        const water = new Water(
            waterGeometry,
            {
                textureWidth: 512,
                textureHeight: 512,
                waterNormals: new THREE.TextureLoader().load("textures/waternormals.jpg", (texture) => {
                    texture.wrapS = texture.wrapT = THREE.RepeatWrapping;
                }),
                sunDirection: new THREE.Vector3(),
                sunColor: 0xffffff,
                waterColor: 0x001e0f,
                distortionScale: 3.7,
                fog: this._scene.fog !== undefined

            }
        );

        water.rotation.x = -Math.PI / 2;
        this._scene.add(water);

        const sky = new Sky();
        sky.scale.setScalar(10000);
        this._scene.add(sky);

        const skyUniforms = sky.material.uniforms;
        skyUniforms["turbidity"].value = 10;
        skyUniforms["rayleigh"].value = 2;
        skyUniforms["mieCoefficient"].value = 0.005;
        skyUniforms["mieDirectionalG"].value = 0.8;

        const parameters = {
            elevation: 2,
            azimuth: 180
        };

        this._parameters = parameters;
        this._sky = sky;
        this._water = water;

        this._updateSun();

        const geometry = new THREE.BoxGeometry(30, 30, 30);
        const material = new THREE.MeshStandardMaterial({ roughness: 0 });

        const mesh = new THREE.Mesh(geometry, material);
        this._scene.add(mesh);
        this._mesh = mesh;
        
    }

    _setupHelper() {
        //.
    }

    _setupGUI() {
        const stats = new Stats();
        document.body.appendChild(stats.dom);
        this._stats = stats;

        const gui = new GUI();

        const folderSky = gui.addFolder("Sky");
        folderSky.add(this._parameters, "elevation", 0, 90, 0.1).onChange(this._updateSun.bind(this));
        folderSky.add(this._parameters, "azimuth", -180, 180, 0.1).onChange(this._updateSun.bind(this));
        folderSky.open();

        const waterUniforms = this._water.material.uniforms;
        const folderWater = gui.addFolder("Water");
        folderWater.add(waterUniforms.distortionScale, "value", 0, 8, 0.1).name("distortionScale");
        folderWater.add(waterUniforms.size, "value", 0.1, 10, 0.1).name("size");
        folderWater.open();
    }

    _setupCamera() {
        const camera = new THREE.PerspectiveCamera(55, window.innerWidth/window.innerHeight, 1, 20000);
        camera.position.set(30, 30, 100);
        this._camera = camera;
    }

    _setupControls() {
        const controls = new OrbitControls(this._camera, this._renderer.domElement);
        controls.maxPolarAngle = Math.PI * 0.495;
        controls.target.set(0, 10, 0);
        controls.minDistance = 40.0;
        controls.maxDistance = 200.0;
        controls.update();
    }

    _setupLight() {
    }

    update() {
        const time = performance.now() * 0.001;
        this._mesh.position.y = Math.sin(time) * 20 + 5;
        this._mesh.rotation.x = time * 0.5;
        this._mesh.rotation.z = time * 0.51;

        this._water.material.uniforms["time"].value += 1.0 / 60.0;

        this._stats.update();
    }

    render() {
        requestAnimationFrame(this.render.bind(this));
        this._renderer.render(this._scene, this._camera);   
        
        this.update();
    }

    resize() {
        const width = window.innerWidth;
        const height = window.innerHeight;

        this._camera.aspect = width / height;
        this._camera.updateProjectionMatrix();
        
        this._renderer.setSize(width, height);
    }
}

window.onload = function () {
    new App();
}